-- Based on LibGuild-1.0

local MAJOR_VERSION = "LibRaid-1.0"
local MINOR_VERSION = tonumber(("$Revision: 81803 $"):match("%d+")) or 0

local lib, oldMinor = LibStub:NewLibrary(MAJOR_VERSION, MINOR_VERSION)
if not lib then return end

local BZ

local GetNumRaidMembers = _G.GetNumRaidMembers
local select = _G.select
local next = _G.next

local new, del, clear, doAll
do
	local cache = {}
	function new(...)
		local t = next(cache)
		if t then
			cache[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return { ... }
		end
	end
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		cache[t] = true
		return nil
	end
	function clear(t)
		for k in pairs(t) do
			t[k] = nil
		end
		return nil
	end
	function doAll(func, ...)
		for i=1, select('#', ...) do
			func(select(i, ...))
		end
	end
end

local frame
if lib.frame then
	frame = lib.frame
	frame:UnregisterAllEvents()
	frame:SetScript("OnEvent", nil)
	frame:SetScript("OnUpdate", nil)
else
	frame = CreateFrame("Frame", MAJOR_VERSION .. "_Frame")
	lib.frame = frame
end

frame:SetScript("OnEvent", function(this, event, ...)
	this[event](lib, ...)
end)

local CallbackHandler = LibStub:GetLibrary("CallbackHandler-1.0")
if lib.callbacks then
	lib:UnregisterAll(lib)
else
	lib.callbacks = LibStub("CallbackHandler-1.0"):New(lib, nil, nil, "UnregisterAll")
end

frame:RegisterEvent("RAID_ROSTER_UPDATE")

local players = {}
local playerRanks = {}
local playerSubgroups = {}
local playerLevels = {}
local playerClasses = {}
local playerClassesEnglish = {}
local playerZones = {}
local playersOnline = {}
local playersDead = {}
local playerRoles = {}
local masterLooter = nil
local numPlayersOnline = 0
local numPlayersTotal = 0
local firstScanDone = false
local raidLeader = 2
local raidAssist = 1

local function IsInRaid()
	return GetNumRaidMembers() > 0
end

local playersOnlineOld, playersOld, playerLevelsOld = {}, {}, {}
function frame:RAID_ROSTER_UPDATE()
	if not IsInRaid() then
		doAll(clear, players, playerRanks, playerSubgroups, playerLevels, playerClasses, playerClassesEnglish, playerZones, playersOnline, playersDead, playerRoles)
		numPlayersTotal = 0
		return
	end
	playersOnline, playersOnlineOld = playersOnlineOld, playersOnline -- for on/offline events
	players, playersOld = playersOld, players -- for add/remove events
	numPlayersTotal = GetNumRaidMembers()
	
	for i = 1, numPlayersTotal do
		local name, rank, subgroup, level, class, englishClass, zone, online, isDead, role, isML = GetRaidRosterInfo(i)
		local add, connect

		if isDead == "" then isDead = nil end
		if name then
			players[name] = true
			playerRanks[name] = rank or UNKNOWN
			playerSubgroups[name] = subgroup
			playerLevels[name] = level or -1 -- note: -1 if offline
			playerClasses[name] = class or UNKNOWN
			playerClassesEnglish[name] = englishClass or UNKNOWN
			playerZones[name] = zone or UNKNOWN
			playersOnline[name] = online
			playersDead[name] = isDead
			playerRoles[name] = role
			if isML == 1 then
				masterLooter = name
			end

			if not playersOld[name] then
				add = true
			end
		end
		if online then
			numPlayersOnline = numPlayersOnline + 1
			if name then
				playersOnline[name] = true
				if playersOnlineOld[name] then
					playersOnlineOld[name] = nil
				elseif playersOld[name] then
					connect = true
				end
			end
		end
		if firstScanDone then
			if add then
				lib.callbacks:Fire("Added", name)
			end
			if connect then
				lib.callbacks:Fire("Connected", name)
			end
		end
	end
	for name in pairs(playersOld) do
		playersOld[name] = nil
		lib.callbacks:Fire("Removed", name)
	end
	for name in pairs(playersOnlineOld) do
		playersOnlineOld[name] = nil
		lib.callbacks:Fire("Disconnected", name)
	end
	if firstScanDone == false then
		lib.callbacks:Fire("Joined")
		firstScanDone = true
	end	
	lib.callbacks:Fire("Update")
end

function lib:GetNumOnline()
	return numPlayersOnline
end

function lib:GetNumOffline()
	return numPlayersTotal - numPlayersOnline
end

function lib:GetNumTotal()
	return numPlayersTotal
end

local playerName = UnitName("player")

function lib:HasMember(name)
	return players[name or playerName]
end

function lib:IsMasterLooter(name)
	return (name or playerName) == masterLooter
end

function lib:IsOnline(name)
	return playersOnline[name or playerName] or false
end

function lib:IsRaidLeader(name)
	return playerRanks[name or playerName] == raidLeader
end

function lib:IsRaidAssist(name)
	return playerRanks[name or playerName] == raidAssist
end

function lib:GetRank(name)
	return playerRanks[name or playerName]
end

function lib:GetSubgroup(name)
	return playerSubgroups[name or playerName]
end

function lib:GetLevel(name)
	return playerLevels[name or playerName]
end

function lib:GetClass(name)
	return playerClasses[name or playerName], playerClassesEnglish[name or playerName]
end

function lib:GetClassColor(name)
	local class = playerClassesEnglish[name or playerName]
	if not class then return 0.8, 0.8, 0.8 end
	local c = RAID_CLASS_COLORS[class]
	return c.r, c.g, c.b
end

function lib:GetClassHexColor(name)
	local class = playerClassesEnglish[name or playerName]
	if not class then return 'cccccc' end
	local c = RAID_CLASS_COLORS[class]
	return ("%02X%02X%02X"):format(c.r*255, c.g*255, c.b*255)
end

function lib:GetZone(name)
	return playerZones[name or playerName]
end

function lib:GetEnglishZone(name)
	if not BZ then BZ = LibStub("LibBabble-Zone-3.0"):GetReverseLookupTable() end
	local zone = playerZones[name]
	if zone ~= UNKNOWN then
		return BZ[zone]
	else
		return "Unknown"
	end
end

function lib:IsDead(name)
	return playersDead[name or playerName]
end

function lib:GetRole(name)
	return playerRoles[name or playerName]
end

function lib:GetMasterLooter()
	return masterLooter
end

function lib:GetPlayers()
	return players
end

function lib:GetPlayersWithClassColor()
	local t = {}
	for k, _ in pairs(players) do
		t[k] = "|cff"..lib:GetClassHexColor(k)..k.."|r"
	end
	return t
end

local sorts; sorts = {
	NAME = function(a,b)
		return a < b
	end,
	LEVEL = function(a,b)
		if playerLevels[a] == playerLevels[b] then
			return sorts.NAME(a,b)
		else
			return playerLevels[a] < playerLevels[b]
		end
	end,
	CLASS = function(a,b)
		if playerClasses[a] == playerClasses[b] then
			return sorts.LEVEL(a,b)
		else
			return playerClasses[a] < playerClasses[b]
		end
	end,
	ZONE = function(a,b)
		if playerZones[a] == playerZones[b] then
			return sorts.CLASS(a,b)
		else
			return playerZones[a] < playerZones[b]
		end
	end,
	RANK = function(a,b)
		if playerRank[a] == playerRank[b] then
			return sorts.CLASS(a,b)
		else
			return playerRank[a] < playerRank[b]
		end
	end,
	ONLINE = function(a,b)
		if playersOnline[a] == playersOnline[b] then
			return sorts.CLASS(a,b)
		else
			return playersOnline[a] --1 or nil, you see.
		end
	end,
}

local iter = function(t)
	local n = t.n + 1
	t.n = n
	return t[n] or del(t)
end

function lib:GetIterator(includeOffline, sort)
	local sortFunc = sorts[sort or "NAME"]
	if not sortFunc then
		error('Argument #2 must be "NAME", "LEVEL", "CLASS", "ZONE", "RANK", "ONLINE", or nil.')
	end
	
	local tmp = new()
	for k in pairs(includeOffline and players or playersOnline) do
		tmp[#tmp+1] = k
	end
	
	table.sort(tmp, sortFunc)
	tmp.n = 0
	
	return iter, tmp, nil
end

if IsInRaid() then
	frame:Show()
else
	frame:Hide()
	firstScanDone = false
	lib.callbacks:Fire("Left")
end
